/*
 * on Windows you can build this with "cl /Oi /O2 rust.cc"
 * on GCC platforms use "g++ -O2 -march=native rust.cc"
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <emmintrin.h>

typedef uint8_t u8;
typedef uint64_t u64;

#if defined( _WIN32 )

#include <windows.h>

inline double get_time() {
        LARGE_INTEGER counter;
        QueryPerformanceCounter( &counter );

        LARGE_INTEGER freq;
        QueryPerformanceFrequency( &freq );

        return double( counter.QuadPart ) / double( freq.QuadPart );
}

#elif defined( __APPLE__ )

#include <mach/mach_time.h>

inline double get_time() {
        mach_timebase_info_data_t freq;
        mach_timebase_info( &freq );

        u64 ticks = mach_absolute_time();

        return double( ticks ) * ( freq.numer / 1e9 ) * double( freq.denom );
}

#else

#include <time.h>

inline double get_time() {
        struct timespec ts;
        clock_gettime( CLOCK_MONOTONIC, &ts );

        return double( ts.tv_sec ) + ts.tv_nsec / 1e9;
}

#endif

#if defined( _MSC_VER )
#define popcnt64 __popcnt64
#else
#define popcnt64 __builtin_popcountll 
#endif

size_t count_slow( const u8 * haystack, size_t n, u8 needle ) {
	size_t c = 0;
	for( size_t i = 0; i < n; i++ ) {
		if( haystack[ i ] == needle ) {
			c++;
		}
	}
	return c;
}

size_t count_fast( const u8 * haystack, size_t n, u8 needle ) {
	const u8 * one_past_end = haystack + n;
	size_t c = 0;

	// peel
	while( uintptr_t( haystack ) % 16 != 0 && haystack < one_past_end ) {
		if( *haystack == needle ) {
			c++;
		}
		haystack++;
	}

	// haystack is now aligned to 16 bytes
	// loop as long as we have 32 bytes left in haystack
	__m128i needles = _mm_set1_epi8( needle );
	while( haystack < one_past_end - 32 ) {
		{
			__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
			haystack += 16;
			__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
			u64 pophi = popcnt64( _mm_cvtsi128_si64( _mm_unpackhi_epi64( cmp, cmp ) ) );
			u64 poplo = popcnt64( _mm_cvtsi128_si64( cmp ) );
			c += ( pophi + poplo ) / 8;
		}
		{
			__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
			haystack += 16;
			__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
			u64 pophi = popcnt64( _mm_cvtsi128_si64( _mm_unpackhi_epi64( cmp, cmp ) ) );
			u64 poplo = popcnt64( _mm_cvtsi128_si64( cmp ) );
			c += ( pophi + poplo ) / 8;
		}
	}

	// remainder
	while( haystack < one_past_end ) {
		if( *haystack == needle ) {
			c++;
		}
		haystack++;
	}

	return c;
}

int main() {
	// build an array of 10000 Lorem Ipsums
	int N = 100000;
	u8 lorem[] = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";

	size_t sz = sizeof( lorem ) * N;
	u8 * lorems = ( u8 * ) malloc( sz );
	if( lorems == NULL ) return 0;

	for( int i = 0; i < N; i++ ) {
		memcpy( lorems + i * sizeof( lorem ), lorem, sizeof( lorem ) );
	}

	// count_slow
	for( int i = 0; i < 10; i++ ) {
		double before = get_time();
		size_t c = count_slow( lorems, sz, 'o' );
		double dt = get_time() - before;
		printf( "count_slow found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}

	// count_fast
	for( int i = 0; i < 10; i++ ) {
		double before = get_time();
		size_t c = count_fast( lorems, sz, 'o' );
		double dt = get_time() - before;
		printf( "count_fast found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}

	// test peeling and remainder
	{
		double before = get_time();
		size_t c = count_fast( lorems + 3, sz - 3, 'o' );
		double dt = get_time() - before;
		printf( "count_fast found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}

	{
		double before = get_time();
		size_t c = count_fast( lorems + 3, sz - 9, 'o' );
		double dt = get_time() - before;
		printf( "count_fast found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}

	// test < 16 elements
	{
		double before = get_time();
		size_t c = count_fast( lorems, 10, 'o' );
		double dt = get_time() - before;
		printf( "count_fast found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}

	return 0;
}
